from __future__ import absolute_import

import argparse
import csv

import sys

from phigaro import const
from phigaro.finder.base import AbstractFinder, Phage
from phigaro.finder.v2 import V2Finder
from phigaro.data import convert_npn


def find_and_print_phages(npn_filename, finder, sep='\t'):
    # type: (str, AbstractFinder) -> list[(str, str, list[Phage])]

    writer = csv.writer(sys.stdout, delimiter='\t')
    writer.writerow(('scaffold', 'begin', 'end'))

    with open(npn_filename) as f:
        reader = csv.reader(f, delimiter=sep)
        for scaffold, npn_str in reader:
            npn = convert_npn(npn_str, 'P')
            phages = finder.find_phages(npn)
            for phage in phages:
                writer.writerow((scaffold, phage.begin, phage.end))


def main():
    p = argparse.ArgumentParser(description="Tool for predicting prophages in metagenomes")
    p.add_argument('npn_file')
    p.add_argument('-w', '--window-len', default=const.DEFAULT_WINDOW_SIZE)
    p.add_argument('-m', '--threshold-min', default=const.DEFAULT_THRESHOLD_MIN)
    p.add_argument('-M', '--threshold-max', default=const.DEFAULT_THRESHOLD_MAX)

    args = p.parse_args()

    finder = V2Finder(
        window_len=args.window_len,
        threshold_min=args.threshold_min,
        threshold_max=args.threshold_max,
    )

    find_and_print_phages(args.npn_file, finder)


if __name__ == '__main__':
    main()



from __future__ import division
from builtins import range

from .base import AbstractFinder, Phage
from functools import partial


class V2Finder(AbstractFinder):
    def __init__(self, window_len, threshold_min, threshold_max):
        """
        :type window_len: int
        :type threshold_min: float
        :type threshold_max: float
        """
        self.window_len = window_len
        self.threshold_min = threshold_min
        self.threshold_max = threshold_max

    def find_phages(self, bacteria_npn):
        scores = calc_scores(bacteria_npn, self.window_len, score_tri)
        ranges = scan_phages(scores, self.threshold_min, self.threshold_max)
        for (begin, end) in ranges:
            yield Phage(
                begin=begin,
                end=end,
                is_prophage=True,
            )


def calc_scores(npn, window_len, score_func):
    def zeros(l):
        return [0] * l
    scores = []
    len_d2 = window_len // 2
    tphage = zeros(len_d2) + npn + zeros(len_d2)
    for i in range(len_d2, len(npn) + len_d2):
        begin = i-len_d2
        end = i+len_d2 + 1
        part = tphage[begin: end]
        scores.append(score_func(part))
    return scores


def plane_kernel(pos):
    return 1


def tri_kernel(pos):
    return 1 - abs(pos - 0.5)/0.5


def score(part, kernel_func):
    return sum(
        c * kernel_func(i/(len(part) - 1))
        for i, c in enumerate(part)
    )


score_tri = partial(score, kernel_func=tri_kernel)
score_plane = partial(score, kernel_func=plane_kernel)


def scan_phages(scores, threshold_min, threshold_max):
    ranges = []

    begin = None
    end = None
    has_peak = False

    scores += [0]
    for i, v in enumerate(scores):
        if v >= threshold_max:
            has_peak = True
        if v >= threshold_min:
            if begin is None:
                begin = i
            end = i
        else:
            if begin is not None and has_peak:
                ranges.append((begin, end))
            begin = None
            end = None
            has_peak = False

    return ranges



import csv
import logging
import re
from builtins import super
from itertools import groupby

from .base import AbstractTask
from .gene_mark import GeneMarkTask
from .hmmer import HmmerTask

logger = logging.getLogger(__name__)


class ParseHmmerTask(AbstractTask):
    task_name = 'parse_hmmer'

    def __init__(self, hmmer_task, gene_mark_task):
        """

        :type hmmer_task: HmmerTask
        :type gene_mark_task: GeneMarkTask
        """
        super().__init__()
        self.hmmer_task = hmmer_task
        self.genemark_task = gene_mark_task

    def output(self):
        return self.file('{}.npn'.format(self.sample))

    def run(self):
        self._parse_hmmer_output()

    @staticmethod
    def parse_line(line):
        tokens = re.split(r'\s+', line)
        scaffold = '>' + line.split('>')[1]
        name = tokens[0]
        evalue = float(tokens[4])
        return scaffold, name, evalue

    def _parse_hmmer_output(self):
        max_evalue = self.config['hmmer']['e_value_threshold']

        with open(self.hmmer_task.output()) as f:
            it = (
                self.parse_line(line.strip())
                for line in f
                if not line.startswith('#') and line.strip()
            )

            hmm_res = {}
            for scaffold, name, evalue in it:
                if scaffold not in hmm_res:
                    hmm_res[scaffold] = {}
                hmm_res[scaffold][name] = 1 if evalue <= max_evalue else 0

        with open(self.genemark_task.output()) as f:
            with open(self.output(), 'w') as of:
                writer = csv.writer(of, delimiter='\t')
                it = (
                    line.strip().split('\t')
                    for line in f
                    if line.startswith('>')
                )

                for scaffold, group in groupby(it, key=lambda t: t[1]):
                    if scaffold not in hmm_res:
                        continue
                    scaffold_res = hmm_res[scaffold]

                    is_phage_it = (
                        scaffold_res.get(name[1:], 0)
                        for name, _ in group
                    )

                    is_phage_it = (
                        'P' if is_phage else 'N'
                        for is_phage in is_phage_it
                    )
                    writer.writerow((scaffold, ''.join(is_phage_it)))




from __future__ import absolute_import

import argparse
import os

import sys
import yaml
from os.path import exists, dirname

from phigaro.helper import SetupHelper, HelperException, download_pvogs


def create_config(no_update_db, config_path, pvogs_dir):
    config_dir = dirname(config_path)
    if not exists(config_dir):
        os.makedirs(config_dir)

    if exists(config_path):
        print('Phigaro already configured')
        return

    helper = SetupHelper(no_update_db)
    try:
        config = helper.setup()
        # TODO: refactor 'allpvoghmms' const
        config['hmmer']['pvog_path'] = os.path.join(pvogs_dir, 'allpvoghmms')

        print("Found MetaGeneMark in: {}".format(config['genemark']['bin']))
        print("Found MetaGeneMark model in: {}".format(config['genemark']['mod_path']))
        print("Found HMMER in: {}".format(config['hmmer']['bin']))
        print("HMMER model in: {}".format(config['hmmer']['pvog_path']))

        with open(config_path, 'w') as f:
            yaml.safe_dump(config, f, default_flow_style=False)
    except HelperException as ex:
        sys.stdout.write(ex.message+'\n')
        exit(1)
        return


def read_config_and_download_pvogs(config_path):
    if not exists(config_path):
        print("You must configure phigaro first")
        exit(1)

    with open(config_path) as f:
        config = yaml.load(f)
        pvogs_dir = dirname(config['hmmer']['pvog_path'])
        print('Downloading models to {}'.format(pvogs_dir))
        download_pvogs('http://download.ripcm.com/phigaro/', pvogs_dir)


def main():
    home = os.getenv('HOME')
    phigaro_dir = os.path.join(home, '.phigaro')
    phigaro_config = os.path.join(phigaro_dir, 'config.yml')
    pvogs_dir = os.path.join(phigaro_dir, 'pvog')

    parser = argparse.ArgumentParser(prog='phigaro-setup',
                                     description="Phigaro setup helper",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config', default=phigaro_config, help='config path')
    parser.add_argument('-p', '--pvog', default=pvogs_dir, help='pvogs dir')
    parser.add_argument('--no-updatedb', action='store_true', help='Do not run sudo updatedb')
    # parser.add_argument('--download-models', action='store_true', help='Skip configuration step and download models')
    args = parser.parse_args()

    create_config(
        no_update_db=args.no_updatedb,
        config_path=args.config,
        pvogs_dir=args.pvog,
    )

    read_config_and_download_pvogs(args.config)

if __name__ == '__main__':
    main()



from __future__ import absolute_import

import argparse
import logging
import multiprocessing
import os
import sys
import uuid
from os.path import join, exists

import yaml

from phigaro.context import Context
from phigaro.scheduling.path import sample_name
from phigaro.scheduling.runner import run_tasks_chain
from phigaro.scheduling.task.gene_mark import GeneMarkTask
from phigaro.scheduling.task.hmmer import HmmerTask
from phigaro.scheduling.task.dummy import DummyTask
from phigaro.scheduling.task.parse_hmmer import ParseHmmerTask
from phigaro.scheduling.task.run_phigaro import RunPhigaroTask
from phigaro._version import __version__

def parse_substitute_output(subs):
    subs = subs or []
    res = {}
    for sub in subs:
        task_name, output = sub.split(":")
        res[task_name] = DummyTask(output, task_name)
    return res


def create_task(substitutions, task_class, *args, **kwargs):
    # TODO: refactor to class Application
    task = task_class(*args, **kwargs)
    if task.task_name in substitutions:
        print('Substituting output for {}: {}'.format(
            task.task_name, substitutions[task.task_name].output()
        ))

        return substitutions[task.task_name]
    return task


def main():
    default_config_path = join(os.getenv('HOME'), '.phigaro', 'config.yml')
    parser = argparse.ArgumentParser(
        prog='phigaro',
        description='Phigaro is a scalable command-line tool for predictions phages and prophages '
                    'from nucleid acid sequences',
    )

    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument('-f', '--fasta-file', help='Assembly scaffolds/contigs or full genomes, required',
                        required=True)
    parser.add_argument('-c', '--config', default=default_config_path, help='Config file, not required')
    parser.add_argument('-v', '--verbose', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-o', '--output', help='Output file, not required, default is stdout')
    parser.add_argument('-t', '--threads',
                        type=int,
                        default=multiprocessing.cpu_count(),
                        help='Num of threads ('
                             'default is num of CPUs={})'.format(multiprocessing.cpu_count()))

    parser.add_argument('--no-cleanup', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-S', '--substitute-output', action='append', help=argparse.SUPPRESS)

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
    logging.getLogger('sh.command').setLevel(logging.WARN)

    logger = logging.getLogger(__name__)

    if not exists(args.config):
        # TODO: pretty message
        print('Please create config file using phigaro-setup script')
        exit(1)

    with open(args.config) as f:
        logger.info('Using config file: {}'.format(args.config))
        config = yaml.load(f)

    filename = args.fasta_file
    sample = '{}-{}'.format(
        sample_name(filename),
        uuid.uuid4().hex
    )

    Context.initialize(
        sample=sample,
        config=config,
        threads=args.threads,
    )

    substitutions = parse_substitute_output(args.substitute_output)

    gene_mark_task = create_task(substitutions,
                                 GeneMarkTask,
                                 filename)
    hmmer_task = create_task(substitutions,
                             HmmerTask,
                             gene_mark_task=gene_mark_task)
    parse_hmmer_task = create_task(substitutions,
                                   ParseHmmerTask,
                                   gene_mark_task=gene_mark_task,
                                   hmmer_task=hmmer_task,
                                   )
    run_phigaro_task = create_task(substitutions,
                                   RunPhigaroTask,
                                   gene_mark_task=gene_mark_task,
                                   parse_hmmer_task=parse_hmmer_task)

    tasks = [
        gene_mark_task,
        hmmer_task,
        parse_hmmer_task,
        run_phigaro_task
    ]
    task_output_file = run_tasks_chain(tasks)

    with open(task_output_file) as f:
        out_f = open(args.output, 'w') if args.output else sys.stdout
        for line in f:
            out_f.write(line)
        if out_f is sys.stdout:
            out_f.close()

    if not args.no_cleanup:
        for t in tasks:
            t.clean()

if __name__ == '__main__':
    main()



from builtins import super

import sh

from .base import AbstractTask
from .gene_mark import GeneMarkTask


class HmmerTask(AbstractTask):
    task_name = 'hmmer'

    def __init__(self, gene_mark_task):
        """

        :type gene_mark_task: GeneMarkTask
        """
        super().__init__()
        self.gene_mark = gene_mark_task

    def _prepare(self):
        self.hmmer = (
            sh.Command(self.config['hmmer']['bin'])
        )

    def output(self):
        return self.file('{}.npn'.format(self.sample))

    def run(self):
        self.hmmer('--cpu', self.context.threads,
                   '--notextw',
                   '--tblout', self.output(),
                   self.config['hmmer']['pvog_path'],
                   self.gene_mark.output())





from os.path import join, basename, exists
from os import makedirs, unlink
import re


def sample_name(sample_path):
    sample_path = basename(sample_path)
    return re.match(r'(\w+)', sample_path).group(1)


def path(*items):
    items = ('proc', ) + items
    return join(*items)


def remove(*items):
    target = path(*items)
    unlink(target)


def directory(*items):
    dir_name = path(*items)
    if not exists(dir_name):
        makedirs(dir_name)
    return dir_name


def file(*items):
    directory(*items[:-1])
    return path(*items)



import abc
import six


class Phage(object):
    def __init__(self, begin, end, is_prophage):
        """

        :type begin: int
        :type end: int
        :type is_prophage: bool
        """
        self.begin = begin
        self.end = end
        self.is_prophage = is_prophage


@six.add_metaclass(abc.ABCMeta)
class AbstractFinder(object):
    @abc.abstractmethod
    def find_phages(self, bacteria_npn):
        """

        :type bacteria_npn: list[int]
        :rtype: list[Phage]
        """
        pass








from builtins import super

from .base import AbstractTask


class DummyTask(AbstractTask):
    task_name = 'dummy'

    def __init__(self, output, old_task_name):
        super().__init__()
        self._output = output
        self.task_name = '{}-dummy'.format(old_task_name)

    def run(self):
        pass

    def output(self):
        return self._output

    def clean(self):
        pass



from builtins import input, str
import os
import time
from os.path import join, dirname, exists
from glob import glob
from shutil import copy
import sh
from future.backports.urllib.parse import urljoin
import re


class HelperException(Exception):
    message = None


class MetaGeneMarkNotFound(HelperException):
    message = "Didn't find MetaGeneMark anywhere, " + \
              "please download it from http://topaz.gatech.edu/Genemark/license_download.cgi "


class MetaGeneMarkKeyNotFound(HelperException):
    message = "Please update your MetaGeneMark key file: " + \
              "http://topaz.gatech.edu/GeneMark/license_download.cgi . " + \
              "You can download the \"key\" file only. "


class HMMERNotFound(HelperException):
    message = "No HMMER found. Please download HMMER package from here: http://hmmer.org/download.html "


def _choose_option(message, options):
    options = [
        s.rstrip()
        for s in options
    ]
    if len(options) == 1:
        return options[0]

    print(message)
    while True:
        res = None

        for i, option in enumerate(options):
            print("[{}] {}".format(i + 1, option))
        option_num = str(input('Choose your option (Enter for {}): '.format(options[0])))
        if option_num == '':
            option_num = '1'
        if re.match(r'^\d+$', option_num):
            option_num = int(option_num) - 1
            if 0 <= option_num < len(options):
                res = options[option_num]
        if res is None:
            print('Please select valid option number ({} to {})'.format(1, len(options)))
            continue
        return res


def download_pvogs(base_url, out_dir):
    def download_file(filename):
        url = urljoin(base_url, filename)
        out = join(out_dir, filename)
        print('Downloading {url} to {out}'.format(
            url=url,
            out=out,
        ))
        sh.wget('-O', out, url, _tty_out=True)

    if not exists(out_dir):
        os.makedirs(out_dir)
    # TODO: refactor 'allpvoghmms' const
    download_file('allpvoghmms')
    download_file('allpvoghmms.h3f')
    download_file('allpvoghmms.h3i')
    download_file('allpvoghmms.h3m')
    download_file('allpvoghmms.h3p')


class SetupHelper(object):
    def __init__(self, dont_update_db):
        self.HOME = os.getenv("HOME")
        self._dont_update_db = dont_update_db

    def _locate(self, *args, **kwargs):
        if not self._dont_update_db:
            try:
                with sh.contrib.sudo:
                    sh.updatedb()
                    self._dont_update_db = True
            except sh.ErrorReturnCode_1:
                print('Invalid password')
                exit(1)
        try:
            return [
                s.rstrip()
                for s in sh.locate(*args, **kwargs)
            ]
        except sh.ErrorReturnCode_1:
            return []

    def _find_binary(self, name, options_message, raise_if_not_found):
        locations = [sh.which(name) or ''] + self._locate('-b', '-r', '^{}$'.format(name))
        locations = [
            l.rstrip()
            for l in locations
            if l.rstrip()
        ]

        if not locations:
            raise raise_if_not_found()

        return _choose_option(
            message=options_message,
            options=locations,
        )

    def _find_mgm_mod_file(self, mgm_dir):
        # TODO: handle no or multiple .mod files in mgm_dir
        message = 'Please select appropriate MetaGeneMark .mod file location'
        res = glob(join(mgm_dir, '*.mod'))
        return _choose_option(message=message,
                              options=res)

    def _is_gm_key_valid(self, gm_key_path):
        res = os.stat(gm_key_path)
        days = (time.time() - res.st_mtime) / (60 * 60 * 24)
        return days < 400

    def _find_gm_key(self, ):
        gm_keys = [
            gm_key_path.rstrip()
            for gm_key_path in self._locate('gm_key')
        ]

        valid_keys = [
            gm_key_path
            for gm_key_path in gm_keys
            if exists(gm_key_path) and self._is_gm_key_valid(gm_key_path)
        ]
        if not valid_keys:
            raise MetaGeneMarkKeyNotFound()
        return valid_keys[0]

    def _setup_mgm(self, ):
        mgm_location = self._find_binary(name='gmhmmp',
                                         options_message='Please select appropriate MetaGeneMark location',
                                         raise_if_not_found=MetaGeneMarkNotFound)
        mod_file = self._find_mgm_mod_file(dirname(mgm_location))
        gm_key_home = join(self.HOME, '.gm_key')

        if not exists(gm_key_home) or not self._is_gm_key_valid(gm_key_home):
            valid_gm_key = self._find_gm_key()
            if valid_gm_key.endswith(".gz"):
                print('Extracting {} to {}'.format(valid_gm_key, gm_key_home))
                sh.gunzip(valid_gm_key, '-c', _out=gm_key_home)
            else:
                print('Copying {} to {}'.format(valid_gm_key, gm_key_home))
                copy(valid_gm_key, gm_key_home)

        return {
            'bin': mgm_location,
            'mod_path': mod_file,
            # 'valid_key': valid_gm_key,
        }

    def _setup_hmmer(self):
        hmmsearch_location = self._find_binary(name='hmmsearch',
                                               options_message='Please select appropriate HMMER location',
                                               raise_if_not_found=HMMERNotFound)
        return {
            'bin': hmmsearch_location,
            'e_value_threshold': 1.0e-5,
        }

    def setup(self):
        mgm_params = self._setup_mgm()
        hmmer_params = self._setup_hmmer()

        return {
            'genemark': {
                'bin': mgm_params['bin'],
                'mod_path': mgm_params['mod_path'],
            },
            'hmmer': {
                'bin': hmmer_params['bin'],
                'e_value_threshold': 1.0e-5,
            },
            'phigaro': {
                'window_len': 34,
                'threshold_min': 5.341108,
                'threshold_max': 7.571429,
            }
        }
